from enum import Enum


class VerificationStatus(Enum):
    """
    A representation of all possible verification status
    """
    ok = 'ok'
    pending = 'pending'
    expired = 'expired'
    revoked = 'revoked'
    failed = 'failed'
    error = 'error'


class StepStatus(Enum):
    """
    A representation of all status of a verification step.
    """
    not_started = 'not_started'
    passed = 'passed'
    pending = 'pending'
    failed = 'failed'
    error = 'error'


class JobStatus(Enum):
    """
    A representation of all possible status of a Job.
    """
    failed = 'failed'
    started = 'started'
    unconfirmed = 'unconfirmed'
    pending = 'pending'
    finished = 'finished'
    scheduled = 'scheduled'



from datetime import datetime
from typing import TypeVar, Callable
from uuid import UUID

from vdx_helper.domain import VerificationStatus, StepStatus, JobStatus
from vdx_helper.models import File, PaginatedResponse, Credential, Job, VerificationResponse, \
    VerificationStepResult, Certificate, Claim, Partner, \
    VerificationReport
from vdx_helper.util import optional_datetime_from_string

T = TypeVar('T')


def json_mapper(json_: dict) -> dict:
    """
    Directly map any json response to a json response.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: The response in json format
    :rtype: dict
    """
    return json_


def get_paginated_mapper(mapper: Callable[[dict], T]) -> Callable[[dict], 'PaginatedResponse[T]']:
    """
    Obtain a mapper method that maps a json response from the Core API into a PaginatedResponse object containing
    a specific object as its items.

    :param mapper: The mapper to use to map the Paginated result's items
    :type mapper: Callable

    :return: A mapping function
    :rtype: Callable
    """
    def paginated_mapper(json_: dict) -> 'PaginatedResponse[T]':
        """
        Maps the json response into a Paginated Response object.

        :param json_: The json obtained from the endpoint call
        :type json_: dict

        :return: A Paginated Response instance
        :rtype: class:`vdx_helper.models.PaginatedResponse`
        """
        return PaginatedResponse(
            page=int(json_["page"]),
            total_pages=int(json_["total_pages"]),
            per_page=int(json_["per_page"]),
            total_items=int(json_["total_items"]),
            items=[mapper(json_item) for json_item in json_["items"]]
        )

    return paginated_mapper


def file_mapper(json_: dict) -> File:
    """
    Maps the json file response into a File object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A File instance
    :rtype: :class:`vdx_helper.models.File`
    """
    return File(
        file_hash=json_["file_hash"],
        file_type=json_["file_type"]
    )


def credential_mapper(json_: dict) -> Credential:
    """
    Maps the json credential response into a Credential object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A Credential instance
    :rtype: :class:`vdx_helper.models.Credential`
    """
    return Credential(
        uid=UUID(json_["uid"]),
        title=json_["title"],
        metadata=json_["metadata"],
        files=[file_mapper(file) for file in json_["files"]],
        credentials=[credential_mapper(credential) for credential in json_["credentials"]],
        upload_date=datetime.fromisoformat(json_["upload_date"]),
        tags=json_["tags"],
        expiry_date=optional_datetime_from_string(json_["expiry_date"])
    )


def job_mapper(json_: dict) -> Job:
    """
    Maps the json job response into a Job object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A Job instance
    :rtype: :class:`vdx_helper.models.Job`
    """
    return Job(
        uid=UUID(json_["uid"]),
        partner=partner_mapper(json_["partner"]),
        chain=json_["chain"],
        tags=json_["tags"],
        status=JobStatus(json_["status"]),
        created_date=optional_datetime_from_string(json_.get("created_date")),
        start_date=optional_datetime_from_string(json_.get("start_date")),
        issued_date=optional_datetime_from_string(json_.get("issued_date")),
        finished_date=optional_datetime_from_string(json_.get("finished_date")),
        failed_date=optional_datetime_from_string(json_.get("failed_date")),
        scheduled_date=optional_datetime_from_string(json_.get("scheduled_date"))
    )


def verification_mapper(json_: dict) -> VerificationResponse:
    """
    Maps the json verification response into a VerificationResponse object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A VerificationResponse instance
    :rtype: :class:`vdx_helper.models.VerificationResponse`
    """
    return VerificationResponse(
        verification={key: verification_step_mapper(step) for key, step in json_["verification"].items()},
        result=verification_report_mapper(json_["result"])
    )


def verification_step_mapper(json_: dict) -> VerificationStepResult:
    """
    Maps the json verification step result response into a Verification Step Result object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A VerificationStepResult instance
    :rtype: :class:`vdx_helper.models.VerificationStepResult`
    """
    return VerificationStepResult(
        name=json_["name"],
        description=json_["description"],
        status=StepStatus(json_["status"])
    )


def partner_mapper(json_: dict) -> Partner:
    """
    Maps the json partner response into a Partner object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A Partner instance
    :rtype: :class:`vdx_helper.models.Partner`
    """
    return Partner(
        **json_
    )


def claim_mapper(json_: dict) -> Claim:
    """
    Maps the json claim response into a Claim object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A Claim instance
    :rtype: :class:`vdx_helper.models.Claim`
    """
    return Claim(
        uid=UUID(json_["uid"]),
        partner=partner_mapper(json_["partner"]),
        credential=credential_mapper(json_["credential"]),
        issued_date=datetime.fromisoformat(json_["issued_date"]),
        signature=json_["signature"]
    )


def certificate_mapper(json_: dict) -> Certificate:
    """
    Maps the json certificate response into a Certificate object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A Certificate instance
    :rtype: :class:`vdx_helper.models.Certificate`
    """
    json_verification = json_.get("last_verification")
    return Certificate(
        certificate=claim_mapper(json_["certificate"]),
        revoked_date=optional_datetime_from_string(json_.get("revoked_date")),
        last_verification=verification_report_mapper(json_verification) if json_verification is not None else None
    )


def verification_report_mapper(json_: dict) -> VerificationReport:
    """
    Maps the json verification report response into a Verification Report object.

    :param json_: The json obtained from the endpoint call
    :type json_: dict

    :return: A VerificationReport instance
    :rtype: :class:`vdx_helper.models.VerificationReport`
    """
    return VerificationReport(
        status=VerificationStatus(json_["status"]),
        timestamp=datetime.fromisoformat(json_["timestamp"])
    )



from datetime import datetime
from typing import NamedTuple, Optional, List, Mapping, Generic, TypeVar, Dict
from uuid import UUID

from vdx_helper.domain import VerificationStatus, StepStatus, JobStatus

T = TypeVar('T')


class Partner(NamedTuple):
    """
    A Partner object returned by the Core API.

    :param id: The partner ID
    :type id: str

    :param name: The name of the partner
    :type name: str
    """
    id: str
    name: str

    def __eq__(self, obj: 'Partner'):
        """
        Overriding of the equals method for simplified comparison between Paginated objects.

        :param obj: The other instance of a paginated object
        :type obj: :class:`PaginatedResponse`

        :return: True if the two objects are equal
        :rtype: bool
        """
        return isinstance(obj, Partner) and obj.id == self.id and obj.name == self.name


class File(NamedTuple):
    """
    Object representing a File returned by the Core API.

    :param file_hash: The file's hash
    :type file_hash: str

    :param file_type: The file `MIME type <https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types>`
    :type file_type: str, optional
    """
    file_hash: str
    file_type: Optional[str]

    def __eq__(self, obj: 'File'):
        """
        Overriding of the equals method for simplified comparison between File objects.

        :param obj: The other instance of a File object
        :type obj: :class:`File`

        :return: True if the two objects are equal
        :rtype: bool
        """
        return isinstance(obj, File) \
               and obj.file_hash == self.file_hash \
               and obj.file_type == self.file_type


class Credential(NamedTuple):
    """
    Object representing a Credential returned by the Core API.

    :param uid: The uid of the credential
    :type uid: :class:`uuid.UUID`

    :param title: The credential title
    :type title: str

    :param metadata: Any additional data in the credential
    :type metadata: dict

    :param files: List of files issued on the credential
    :type files: List[:class:`File`]

    :param credentials: List of credentials associated to the credential
    :type credentials: List[:class:`Credential`]

    :param upload_date: The date the credential was uploaded
    :type upload_date: :class:`datetime.datetime`

    :param tags: List of tags identifying the credential
    :type tags: List[str]

    :param expiry_date: The date the credential expires, if set
    :type expiry_date: :class:`datetime.datetime`, optional
    """
    uid: UUID
    title: str
    metadata: dict
    files: List[File]
    credentials: List['Credential']
    upload_date: datetime
    tags: List[str]
    expiry_date: Optional[datetime]

    def __eq__(self, obj: 'Credential'):
        """
        Overriding of the equals method for simplified comparison between Credential objects.

        :param obj: The other instance of a credential object
        :type obj: :class:`Credential`

        :return: True if the two objects are equal
        :rtype: bool
        """
        return (
            isinstance(obj, Credential)
            and obj.uid == self.uid
            and obj.title == self.title
            and obj.metadata == self.metadata
            and all(
                first == second for first, second in zip(obj.files, self.files)
            )
            and all(
                first == second
                for first, second in zip(obj.credentials, self.credentials)
            )
            and obj.upload_date == self.upload_date
            and obj.tags == self.tags
            and obj.expiry_date == self.expiry_date
        )


class Job(NamedTuple):
    """
    Object representing a Certificate returned by the Core API.

    :param uid: uid of the Job
    :type uid: :class:`uuid.UUID`

    :param partner: The partner who issued the job
    :type partner: :class:`Partner`

    :param chain: The Blockchain engine the job was issued on
    :type chain: str

    :param tags: List of tags identifying the credential
    :type tags: List[str]

    :param status: Current status of the job
    :type status: :class:`vdx_helper.JobStatus`

    :param created_date: The date the job was created
    :type created_date: :class:`datetime.datetime`

    :param scheduled_date: The date the job will be issued, if not yet issued
    :type scheduled_date: :class:`datetime.datetime`, optional

    :param start_date: The date the job started issuing
    :type start_date: :class:`datetime.datetime`, optional

    :param issued_date: The date the job was issued
    :type issued_date: :class:`datetime.datetime`, optional

    :param finished_date: The date the job issuing was confirmed on the blockchain
    :type finished_date: :class:`datetime.datetime`, optional

    :param failed_date: The date the job issuing failed, if applicable
    :type failed_date: :class:`datetime.datetime`, optional
    """
    uid: UUID
    partner: Partner
    chain: str
    tags: List[str]
    status: JobStatus
    created_date: datetime
    scheduled_date: Optional[datetime] = None
    start_date: Optional[datetime] = None
    issued_date: Optional[datetime] = None
    finished_date: Optional[datetime] = None
    failed_date: Optional[datetime] = None


class VerificationStepResult(NamedTuple):
    """
    Object representing the result of a step of the verification process.

    :param name: The name of the step
    :type name: str

    :param description: The result of the verification step, containing several fields and values
    :type description: dict

    :param status: The final status of the step
    :type status: :class:`vdx_helper.StepStatus`
    """
    name: str
    description: Mapping[str, Optional[str]]
    status: StepStatus


class VerificationReport(NamedTuple):
    """
    Object representing the final result of the verification of a certificate, not including the individual result of
    each step.

    :param status: The status of the verification
    :type status: :class:`vdx_helper.domain.VerificationStatus`

    :param timestamp: The date of the verification
    :type timestamp: :class:`datetime.datetime`
    """
    status: VerificationStatus
    timestamp: datetime


class VerificationResponse(NamedTuple):
    """
    Object representing the result of a verification returned from the API, including all steps.

    :param verification: The list of all verification steps
    :type verification: Dict[str, :class:`VerificationStepResult`]

    :param result: The result of the verification
    :type result: :class:`VerificationReport`
    """
    verification: Dict[str, VerificationStepResult]
    result: VerificationReport


class Claim(NamedTuple):
    """
    Object representing a Credential returned by the Core API.
    A Claim has all data directly related to the issued Certificate.

    :param uid: The uid of the certificate
    :type uid: :class:`uuid.UUID`

    :param partner: The partner who issued the certificate
    :type partner: :class:`Partner`

    :param credential: The credential that was issued
    :type credential: :class:`Credential`

    :param issued_date: Date of issuing on the Blockchain
    :type issued_date: :class:`datetime.datetime`

    :param signature: Signed content of the credential by the partner's issuing key
    :type signature: str
    """
    uid: UUID
    partner: Partner
    credential: Credential
    issued_date: datetime
    signature: str

    def __eq__(self, obj: 'Claim'):
        """
        Overriding of the equals method for simplified comparison between Claim objects.

        :param obj: The other instance of a claim object
        :type obj: :class:`Claim`

        :return: True if the two objects are equal
        :rtype: bool
        """
        return isinstance(obj, Claim) and obj.uid == self.uid


class Certificate(NamedTuple):
    """
    Object representing a Certificate returned by the Core API.

    :param certificate: The details and data of the certificate and respective credential
    :type certificate: :class:`Claim`

    :param revoked_date: The date of revocation, if applicable
    :type revoked_date: :class:`datetime.datetime`, optional

    :param last_verification: The result of the latest verification of the certificate
    :type last_verification: :class:`VerificationReport`
    """
    certificate: Claim
    revoked_date: Optional[datetime]
    last_verification: Optional[VerificationReport]


class PaginatedResponse(NamedTuple, Generic[T]):
    """
    Object representing a paginated response from the API.
    from the API

    :param page: The page the items are from
    :type page: int

    :param total_pages: The total number of pages
    :type total_pages: int

    :param per_page: The number of items returned per page
    :type per_page: int

    :param total_items: The total number of items
    :type total_items: int

    :param items: The list of objects returned by the API
    :type items: List[T]
    """
    page: int
    total_pages: int
    per_page: int
    total_items: int
    items: List[T]

    def __eq__(self, obj: 'PaginatedResponse'):
        """
        Overriding of the equals method for simplified comparison between Paginated objects.

        :param obj: The other instance of a paginated object
        :type obj: :class:`PaginatedResponse`

        :return: True if the two objects are equal
        :rtype: bool
        """
        return (
                isinstance(obj, PaginatedResponse)
                and obj.page == self.page
                and obj.total_pages == self.total_pages
                and obj.per_page == self.per_page
                and obj.total_items == self.total_items
                and all(first == second for first, second in zip(obj.items, self.items))
        )



from http import HTTPStatus

from requests import Response


class VDXError(Exception):
    """
    An exception class representing any error returned by the VDX Core API. It contains the respective
    status code as well as the corresponding error message

    :param code: The status code
    :type code: :class:`http.HTTPStatus`

    :param message: The error message
    :type message: str
    """

    def __init__(self, code: HTTPStatus, message: str):
        self.code = code
        self.message = message


def error_from_response(status: HTTPStatus, response: Response):
    """
    Maps a given error from an endpoint response to an exception with the correct status code and description of the
    error

    :param status: The status code of the endpoint response
    :type status: HTTPStatus

    :param response: The endpoint response
    :type response: :class:`requests.Response`

    :return: A VDXError to be raised
    :rtype: :class:`VDXError`
    """
    if status is not HTTPStatus.OK:
        try:
            json_response = response.json()
            description = json_response.get("description")
        except (ValueError, AttributeError):
            description = ""
        return VDXError(code=status, message=description)



import io
import time
from datetime import datetime
from http import HTTPStatus
from typing import Optional, Callable, Dict, TypeVar, Tuple, List, Iterable, BinaryIO
from uuid import UUID

import requests
from nndict import nndict
from vdx_helper.errors import error_from_response, VDXError
from vdx_helper.mappers import file_mapper, get_paginated_mapper, credential_mapper, job_mapper, \
    certificate_mapper, verification_mapper
from vdx_helper.util import optional_uuid_to_string, optional_uuids_to_string, datetime_from_string, uuids_to_string

T = TypeVar('T')


class VDXHelper:
    """
    Helper class that allows its users to connect to the VDX Core API with little fuss. Each method corresponds to
    an endpoint on the Core API, including all of its parameters. By default, the results are mapped into specific
    objects, but a JSON mapper can also be used (directly returns the results in JSON), as well as custom mappers.
    More information on mappers on the :mod:`vdx_helper.mappers` module.

    To use, create an instance of this class by providing the required information, and then invoke any of the
    available methods. The authentication flow is fully dealt with by this class, as long as the correct client secret
    and URL is provided.

    :param api_url: The url for the VDX Core API. For example, https://vizidox.com/api
    :type api_url: str

    :param auth_url: The url for the VDX Core API authentication server. For example, https://vizidox.com/auth
    :type auth_url: str

    :param client_secret: The client secret for authentication
    :type client_secret: str

    :param client_id: The client ID for authentication
    :type client_id: str
    """

    def __init__(self, api_url: str, auth_url: str, client_secret: str, client_id: str) -> None:
        self.api_url = api_url.rstrip("/")
        self.auth_url = auth_url
        self.client_secret: str = client_secret
        self.client_id: str = client_id

        self.auth_token: Optional[str] = None
        self.token_expiration_date: float = 0

    def _fetch_token(self) -> Tuple[str, float]:
        """
        Retrieves a usable authentication token from the Vizidox Authentication Server, with the set-up
        client_id and client_secret.

        :raises VDXError: Raised when there is an issue with the request to the Authentication server

        :return: A tuple containing the token and its expiration date
        :rtype: Tuple[str, float]
        """
        headers = {
            "Content-Type": "application/x-www-form-urlencoded"
        }

        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "client_credentials"
        }

        response = requests.post(f"{self.auth_url}", headers=headers, data=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise VDXError(status, "API Authentication failed")

        access_token = str(response.json()['access_token'])
        token_expiration_date = time.time() + int(response.json()['expires_in'])

        return access_token, token_expiration_date

    @property
    def _token(self) -> str:
        """
        Checks if there is a token stored in the helper, and if it is expired. If there is no token or if it is
        expired, then a new one is retrieved from the authentication server. Finally, the currently valid token
        is returned.

        :return: The authentication token
        :rtype: str
        """
        if self.auth_token is None or time.time() > self.token_expiration_date:
            self.auth_token, token_expiration_date = self._fetch_token()
            self.token_expiration_date = token_expiration_date

        return self.auth_token

    @property
    def header(self) -> dict:
        """
        Creates a header structure to be used by all requests for authentication.

        :return: The request header
        :rtype: dict
        """
        return {"Authorization": f"Bearer {self._token}", "Accept": "application/json"}

    def upload_file(self,
                    file_stream: BinaryIO,
                    ignore_duplicated: bool = False,
                    mapper: Callable[[dict], T] = file_mapper) -> T:
        """
        Upload a file to the Core API servers.

        :param file_stream: File to be uploaded
        :type file_stream: BinaryIO

        :param ignore_duplicated: Flag indicating if duplicate files should be ignored. If true, and the file has
                                  already been uploaded previously, it will be re-uploaded
        :type ignore_duplicated: bool

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        file_stream.seek(0)
        payload = {"file": (file_stream)}
        form_data = {"ignore_duplicated": ignore_duplicated}

        response = requests.post(f"{self.api_url}/files", headers=self.header, files=payload, data=form_data)

        status = HTTPStatus(response.status_code)
        if status not in [HTTPStatus.OK, HTTPStatus.CREATED]:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_file(self, file_hash: str, mapper: Callable[[dict], T] = file_mapper) -> T:
        """
        Retrieve file details by its hash.

        :param file_hash: File hash to retrieve
        :type file_hash: str

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        response = requests.get(f"{self.api_url}/files/{file_hash}", headers=self.header)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_files(self, mapper: Callable[[dict], T] = get_paginated_mapper(file_mapper),
                  file_hash: Optional[str] = None, **pagination) -> T:
        """
        Retrieve all files uploaded by the Partner, or filter them via their hash.

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param file_hash: File hash to filter by
        :type file_hash: str, optional

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """

        params = nndict(file_hash=file_hash, **pagination)
        response = requests.get(f"{self.api_url}/files", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)
        return mapper(response.json()["result"])

    def get_credentials(self, mapper: Callable[[dict], T] = get_paginated_mapper(credential_mapper), *,
                        metadata: Optional[dict] = None, uid: Optional[UUID] = None,
                        start_date: Optional[datetime] = None, end_date: Optional[datetime] = None,
                        and_tags: Optional[str] = None, or_tags: Optional[str] = None, **pagination) -> T:
        """
        Retrieve all partner credentials, or filter by metadata values, uid,
        upload_date (start_date < upload_date < end_date), and tags.

        More information on tags on the `VDX Core API <https://docs.vizidox.com/#tags>`__ official documentation.
        Results are paginated, and the pagination parameters should be provided as keyword arguments.

        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param metadata: Metadata values to filter by. A dictionary is expected with the key corresponding to the
                         field name found in the metadata, and the value corresponding to that field's value.
        :type metadata: dict, optional

        :param uid: Credential UID to filter by
        :type uid: :class:`uuid.UUID`, optional

        :param start_date: Filter by Credentials uploaded after the given start_date
        :type start_date: :class:`datetime.datetime`, optional

        :param end_date: Filter by Credentials uploaded before the given end_date
        :type end_date: :class:`datetime.datetime`, optional

        :param and_tags: Filter Credentials that contain all the given tags
        :type and_tags: List[str], optional

        :param or_tags: Filter Credentials that contain at least one of the given tags
        :type or_tags: List[str], optional

        :param pagination: Keyword arguments for pagination
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        params = nndict(
            uid=optional_uuid_to_string(uid),
            upload_date_from=start_date,
            upload_date_until=end_date,
            and_tags=and_tags,
            or_tags=or_tags
        )

        metadata_filters = {key if key.startswith('metadata_') else f"metadata_{key}": value
                            for key, value in nndict(metadata)}
        params = {**params, **metadata_filters, **nndict(pagination)}

        response = requests.get(f"{self.api_url}/credentials", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_credential(self, cred_uid: UUID, mapper: Callable[[dict], T] = credential_mapper) -> T:
        """
        Retrieve a specific credential from the Core API, via its UUID.

        :param cred_uid: Credential UID to obtain
        :type cred_uid: :class:`uuid.UUID`

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        response = requests.get(f"{self.api_url}/credentials/{cred_uid}", headers=self.header)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def create_credential(self, title: str, metadata: dict, tags: Optional[Iterable[str]] = None,
                          file_hashes: Optional[List[str]] = None, cred_ids: List[UUID] = None,
                          expiry_date: Optional[str] = None, mapper: Callable[[dict], T] = credential_mapper) -> T:
        """
        Create a new credential from a previously uploaded file, or provided metadata.

        :param title: The title of the credential
        :type title: str

        :param metadata: Metadata values to filter by. A dictionary is expected with the key corresponding to the
                         field name found in the metadata, and the value corresponding to that field's value. Can be
                         an empty dictionary as long as at least one file hash is provided.
        :type metadata: dict

        :param tags: Optional text tags that can be used to identify and filter the credential. Must have at least
                    3 characters, and can only contain alphanumeric characters, '-' or '_'
        :type tags: Iterable[str], optional

        :param file_hashes: List of hashes of files to associate to the credential. Must be hashes of files previously
                            uploaded to the core API. Can be None as long as provided metadata is not empty.
        :type file_hashes: List[str], optional

        :param cred_ids: List of credentials to associate to the new credential.
        :type cred_ids: List[:class:`uuid.UUID`], optional

        :param expiry_date: Date the credential should expire, if applicable
        :type expiry_date: :class:`datetime.datetime`, optional

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        payload = nndict(
            title=title,
            files=file_hashes,
            credentials=optional_uuids_to_string(cred_ids),
            tags=list(set(tags)) if tags is not None else None,
            expiry_date=expiry_date
        )
        payload_json = {**payload, "metadata": dict(metadata)}
        response = requests.post(f"{self.api_url}/credentials", headers=self.header, json=payload_json)

        status = HTTPStatus(response.status_code)
        if status not in [HTTPStatus.OK, HTTPStatus.CREATED]:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def update_credential_tags(self, updated_credential_tags: Iterable[Dict[str, List[str]]]) -> None:
        """
        Add new tags to one or more credentials. Previous credential tags are not changed.

        Example format of the parameter:

        .. code-block:: json

            [
                {
                    "credential_uid": "81fe5e08-46e4-11ec-81d3-0242ac130003",
                    "tags": ["tagA","tagB","tagC"]
                },
                {
                    "credential_uid": "7f026924-b3f8-4670-bc65-37fd7da8867c",
                    "tags": ["tagD"]
                }
            ]

        :param updated_credential_tags: Dictionary containing the credential UUIDs to update and the new tags to add
        :type updated_credential_tags: dict
        """
        payload = {"credentials": updated_credential_tags}
        response = requests.patch(f"{self.api_url}/credentials", headers=self.header, json=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return

    def replace_credential_tags(self, replace_credential_tags: Iterable[Dict[str, List[str]]]) -> None:
        """
        Replaces all the given credential tags with the new ones.

        Example format of the parameter:

        .. code-block:: json

            [
                {
                    "credential_uid": "81fe5e08-46e4-11ec-81d3-0242ac130003",
                    "tags": ["tagA","tagB","tagC"]
                },
                {
                    "credential_uid": "7f026924-b3f8-4670-bc65-37fd7da8867c",
                    "tags": ["tagD"]
                }
            ]

        :param replace_credential_tags: Dictionary containing the credential UUIDs to update and the new tags to replace
                                        the old ones with
        :type replace_credential_tags: dict
        """
        payload = {"credentials": replace_credential_tags}
        response = requests.put(f"{self.api_url}/credentials", headers=self.header, json=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return

    def delete_credential_tag(self, cred_uid: UUID, tag: str) -> None:
        """
        Deletes a specific tag from the given credential.

        :param cred_uid: UUID of the credential to update
        :type cred_uid: :class:`uuid.UUID`

        :param tag: Tag to delete
        :type tag: str
        """
        params = nndict(tag=tag)
        response = requests.patch(f"{self.api_url}/credentials/{cred_uid}/delete_tag",
                                  headers=self.header,
                                  params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

    def schedule_credentials(self, engine: str, credentials: List[UUID], mapper: Callable[[dict], T] = job_mapper) -> T:
        """
        Schedule the given list of credentials to be issued on the Blockchain engine.

        :param engine: Blockchain Engine to schedule the credentials
        :type engine: str

        :param credentials: List of UUIDs of the credentials to schedule
        :type credentials: List[:class:`uuid.UUID`]

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        payload = {
            "engine": engine,
            "credentials": uuids_to_string(credentials)
        }
        response = requests.post(f"{self.api_url}/credentials/schedule", headers=self.header, json=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.CREATED:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def issue_job(self, engine: str, mapper: Callable[[dict], T] = job_mapper) -> T:
        """
        Immediately issues the next scheduled job for the given blockchain engine, if there are scheduled credentials.

        :param engine: Blockchain engine to issue
        :type engine: str

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        payload = {"engine": engine}

        response = requests.post(f"{self.api_url}/jobs/immediate", headers=self.header, json=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.CREATED:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_jobs(self, mapper: Callable[[dict], T] = get_paginated_mapper(job_mapper), *,
                 job_status: Optional[str] = None, uid: Optional[UUID] = None,
                 start_date: Optional[datetime] = None, end_date: Optional[datetime] = None,
                 and_tags: Optional[str] = None, or_tags: Optional[str] = None, **pagination) -> T:
        """
        Retrieve all partner jobs, or filter by a specific status, uid,
        issued date (start_date < issued_date < end_date), and tags.

        More information on tags on the `VDX Core API <https://docs.vizidox.com/#tags>`__ official documentation.
        Results are paginated, and the pagination parameters should be provided as keyword arguments.
        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param job_status: Filter jobs with the given status
        :type job_status: str, optional

        :param uid: Job UID to filter by
        :type uid: :class:`uuid.UUID`, optional

        :param start_date: Filter Jobs that were issued after the given start_date
        :type start_date: :class:`datetime.datetime`, optional

        :param end_date: Filter Jobs that were issued before the given end_date
        :type end_date: :class:`datetime.datetime`, optional

        :param and_tags: Filter Jobs that contain all the given tags
        :type and_tags: List[str], optional

        :param or_tags: Filter Jobs that contain at least one of the given tags
        :type or_tags: List[str], optional

        :param pagination: Keyword arguments for pagination
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        params = nndict(
            uid=optional_uuid_to_string(uid),
            status=job_status,
            issued_date_from=start_date,
            issued_date_until=end_date,
            and_tags=and_tags,
            or_tags=or_tags
        )

        params = {**params, **nndict(pagination)}
        response = requests.get(f"{self.api_url}/jobs", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_job(self, job_uid: UUID, mapper: Callable[[dict], T] = job_mapper) -> T:
        """
        Retrieve a specific job from the Core API, via its UUID.

        :param job_uid: Job UID to obtain
        :type job_uid: :class:`uuid.UUID`

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        response = requests.get(f"{self.api_url}/jobs/{job_uid}", headers=self.header)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def update_job_tags(self, updated_job_tags: List[dict]) -> None:
        """
        Add new tags to one or more jobs. Previous job tags are not changed.

        Example format of the parameter:

        .. code-block:: json

            [
                {
                    "job_uid": "81fe5e08-46e4-11ec-81d3-0242ac130003",
                    "tags": ["tagA","tagB","tagC"]
                },
                {
                    "job_uid": "7f026924-b3f8-4670-bc65-37fd7da8867c",
                    "tags": ["tagD"]
                }
            ]

        :param updated_job_tags: Dictionary containing the job UUIDs to update and the new tags to add
        :type updated_job_tags: dict
        """
        payload = {"jobs": updated_job_tags}
        response = requests.patch(f"{self.api_url}/jobs", headers=self.header, json=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return None

    def replace_job_tags(self, replace_job_tags: List[dict]) -> None:
        """
        Replaces all the given job tags with the new ones.

        Example format of the parameter:

        .. code-block:: json

            [
                {
                    "job_uid": "81fe5e08-46e4-11ec-81d3-0242ac130003",
                    "tags": ["tagA","tagB","tagC"]
                },
                {
                    "job_uid": "7f026924-b3f8-4670-bc65-37fd7da8867c",
                    "tags": ["tagD"]
                }
            ]

        :param replace_job_tags: Dictionary containing the job UUIDs to update and the new tags to replace the old ones with
        :type replace_job_tags: dict
        """
        payload = {"jobs": replace_job_tags}
        response = requests.put(f"{self.api_url}/jobs", headers=self.header, json=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return None

    def get_job_certificates(self, job_uid: UUID, and_tags: Optional[str] = None, or_tags: Optional[str] = None,
                             mapper: Callable[[dict], T] = get_paginated_mapper(certificate_mapper), **pagination) -> T:
        """
        Retrieve all certificates issued in a specific Job, or filter by tags.

        More information on tags on the `VDX Core API <https://docs.vizidox.com/#tags>`__ official documentation.

        Results are paginated, and the pagination parameters should be provided as keyword arguments.

        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param job_uid: UID of the job
        :type job_uid: :class:`uuid.UUID`

        :param and_tags: Filter Certificates issued from credentials that contain all the given tags
        :type and_tags: List[str], optional

        :param or_tags: Filter Certificates issued from credentials that contain at least one of the given tags
        :type or_tags: List[str], optional

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param pagination: Keyword arguments for pagination
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        params = nndict(and_tags=and_tags, or_tags=or_tags)
        params = {**params, **nndict(pagination)}
        response = requests.get(f"{self.api_url}/jobs/{job_uid}/certificates", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_job_credentials(self, job_uid: UUID, and_tags: Optional[str] = None, or_tags: Optional[str] = None,
                            mapper: Callable[[dict], T] = get_paginated_mapper(credential_mapper),
                            **pagination) -> Optional[T]:
        """
        Retrieve all credentials issued in a specific Job, or filter by tags.

        More information on tags on the `VDX Core API <https://docs.vizidox.com/#tags>`__ official documentation.

        Results are paginated, and the pagination parameters should be provided as keyword arguments.

        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param job_uid: UID of the job
        :type job_uid: :class:`uuid.UUID`

        :param and_tags: Filter Credentials that contain all the given tags
        :type and_tags: List[str], optional

        :param or_tags: Filter Credentials that contain at least one of the given tags
        :type or_tags: List[str], optional

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param pagination: Keyword arguments for pagination
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        params = nndict(and_tags=and_tags, or_tags=or_tags)
        params = {**params, **nndict(pagination)}
        response = requests.get(f"{self.api_url}/jobs/{job_uid}/credentials", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def get_certificates(self, mapper: Callable[[dict], T] = get_paginated_mapper(certificate_mapper), *,
                         job_uid: Optional[UUID] = None, cred_uid: Optional[UUID] = None, uid: Optional[UUID] = None,
                         start_date: Optional[datetime] = None, end_date: Optional[datetime] = None,
                         and_credential_tags: Optional[str] = None, or_credential_tags: Optional[str] = None,
                         and_job_tags: Optional[str] = None, or_job_tags: Optional[str] = None,
                         verification_status: Optional[str] = None, **pagination) -> T:
        """
        Retrieve all partner certificates, or filter by job, credential, uid,
        issued date (start_date < issued_date < end_date), and credential or job tags.

        More information on tags on the `VDX Core API <https://docs.vizidox.com/#tags>`__ official documentation.

        Results are paginated, and the pagination parameters should be provided as keyword arguments.

        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param job_uid: Job UID to filter by
        :type job_uid: :class:`uuid.UUID`, optional

        :param cred_uid: Credential UID to filter by
        :type uid: :class:`uuid.UUID`, optional

        :param uid: Certificate UID to filter by
        :type uid: :class:`uuid.UUID`, optional

        :param start_date: Filter by Certificates issued after the given start_date
        :type start_date: :class:`datetime.datetime`, optional

        :param end_date: Filter by Certificates issued before the given end_date
        :type end_date: :class:`datetime.datetime`, optional

        :param and_credential_tags: Filter Certificates of Credentials that contain all the given tags
        :type and_credential_tags: List[str], optional

        :param or_credential_tags: Filter Certificates of Credentials that contain at least one of the given tags
        :type or_credential_tags: List[str], optional

        :param and_job_tags: Filter Certificates issued on Jobs that contain all the given tags
        :type and_job_tags: List[str], optional

        :param or_job_tags: Filter Certificates issued on Jobs that contain at least one of the given tags
        :type or_job_tags: List[str], optional

        :param verification_status: Filter Certificates in the given verification status
        :type verification_status: str, optional

        :param pagination: Keyword arguments for pagination
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        params = nndict(
            uid=optional_uuid_to_string(uid),
            job_uid=optional_uuid_to_string(job_uid),
            credential_uid=optional_uuid_to_string(cred_uid),
            issued_date_from=start_date,
            issued_date_until=end_date,
            and_credential_tags=and_credential_tags,
            or_credential_tags=or_credential_tags,
            and_job_tags=and_job_tags,
            or_job_tags=or_job_tags,
            verification_status=verification_status
        )
        params = {**params, **nndict(pagination)}
        response = requests.get(f"{self.api_url}/certificates", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def download_certificate(self, cert_uid: UUID) -> io.BytesIO:
        """
        Downloads the JSON proof file for the Certificate, containing all relevant Blockchain details as well
        as the metadata of the corresponding issued credential.

        :param cert_uid: UUID of the certificate to download
        :type cert_uid: :class:`uuid.UUID`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        response = requests.get(f"{self.api_url}/certificates/{cert_uid}/download",headers=self.header)

        status = HTTPStatus(response.status_code)

        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return io.BytesIO(response.content)

    def verify_by_uid(self, cert_uid: UUID,
                      mapper: Callable[[dict], T] = get_paginated_mapper(verification_mapper)) -> T:
        """
        Verify a certificate via its UID.

        :param cert_uid: UUID of the certificate to verify
        :type cert_uid: :class:`uuid.UUID`

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        response = requests.get(f"{self.api_url}/verify/{cert_uid}", headers=self.header)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def verify_by_certificate(self, filename: str, file_stream: BinaryIO, file_content_type: str,
                              mapper: Callable[[dict], T] = verification_mapper) -> T:
        """
        Verify a certificate via its proof file.

        :param filename: Name of the file
        :type filename: str

        :param file_stream: Certificate proof file stream
        :type file_stream: BinaryIO

        :param file_content_type: Content type of the file
        :type file_content_type: str

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        file_stream.seek(0)
        payload = {"file": (filename, file_stream, file_content_type)}

        response = requests.post(f"{self.api_url}/verify/upload/certificate", headers=self.header, files=payload)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def verify_by_file(self, filename: str, file_stream: BinaryIO, file_content_type: str,
                       mapper: Callable[[dict], T] = get_paginated_mapper(verification_mapper), **pagination) -> T:
        """
        Verify a certificate via the credential file.
        Results are paginated, and the pagination parameters should be provided as keyword arguments.

        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param filename: Name of the file
        :type filename: str

        :param file_stream: Certificate file stream
        :type file_stream: BinaryIO

        :param file_content_type: Content type of the file
        :type file_content_type: str

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param pagination: Keyword argument containing the pagination parameters
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        payload = {"file": (filename, file_stream, file_content_type)}
        params = {**nndict(pagination)}

        response = requests.post(f"{self.api_url}/verify/upload/file", headers=self.header, files=payload, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def verify_by_credential_uid(self, cred_uid: UUID,
                                 mapper: Callable[[dict], T] = get_paginated_mapper(verification_mapper),
                                 **pagination) -> T:
        """
        Verify certificates via their credential UID.
        Results are paginated, and the pagination parameters should be provided as keyword arguments.

        For more information on the possible parameters, check the
        `VDX Core API <https://docs.vizidox.com/#pagination>`__ official documentation.

        :param cred_uid: UUID of the credential
        :type cred_uid: :class:`uuid.UUID`

        :param mapper: Optional mapper to change the format of the endpoint response
        :type mapper: :class:`typing.Callable`

        :param pagination: Keyword argument containing the pagination parameters
        :type pagination: dict

        :return: The result of the endpoint call
        :rtype: :class:`T`
        """
        params = {**nndict(pagination)}
        response = requests.get(f"{self.api_url}/verify/credential/{cred_uid}", headers=self.header, params=params)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return mapper(response.json()["result"])

    def revoke_certificate(self, cert_uid: UUID) -> datetime:
        """
        Revokes a certificate via its UID, making it no longer valid.

        :param cert_uid: UID of the certificate to be revoked
        :type cert_uid: :class:`uuid.UUID`

        :return: The date the certificate was revoked at
        :rtype: :class:`datetime.datetime`
        """
        response = requests.post(f"{self.api_url}/certificates/{cert_uid}/revoke", headers=self.header)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return datetime_from_string(response.json()["result"])

    def revoke_certificate_by_credential(self, cred_uid: UUID, engine: str) -> datetime:
        """
        Revokes a certificate via the credentials UID and for a specific engine, making it no longer valid.

        :param cred_uid: UID of the credential to be revoked
        :type cred_uid: :class:`uuid.UUID`

        :param engine: Blockchain engine to revoke the credential on
        :type engine: str

        :return: The date the certificate was revoked at
        :rtype: :class:`datetime.datetime`
        """
        response = requests.post(f"{self.api_url}/credentials/{cred_uid}/revoke/{engine}", headers=self.header)

        status = HTTPStatus(response.status_code)
        if status is not HTTPStatus.OK:
            raise error_from_response(status, response)

        return datetime_from_string(response.json()["result"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vdx_helper']

package_data = \
{'': ['*']}

install_requires = \
['nndict==1.0.0', 'requests>=2.27,<3.0']

setup_kwargs = {
    'name': 'vdx-helper',
    'version': '3.2.2',
    'description': 'Python client library to connect to the VDX Core API',
    'long_description': "[![PyPI version](https://badge.fury.io/py/vdx-helper.svg)](https://badge.fury.io/py/vdx-helper) \n[![Downloads](https://static.pepy.tech/personalized-badge/vdx-helper?period=total&units=abbreviation&left_color=black&right_color=blue&left_text=Downloads)](https://pepy.tech/project/vdx-helper)\n\n# VDX Helper\nThis repository provides a wrapper for every call made to VDX Core Api.\n\n## How it works\nThis helper first needs to be authorized by wielding valid token from the authentication server, then use that token for further requests as long as it has not expired.\n\nEach method also allows one to include their own custom mappers, enabling the method to return the result in the format the user wishes.\n\n## Prerequisites\n\n- Python Poetry\n- Docker + Docker-compose\n\n## Usage\n\n### Initialization\n\nRequired parameters: \n- api_url: The url leading to Core API server\n- auth_url: The url leading to authentication server\n- client_secret: The authentication secret\n- client_id: The ID of the client / partner\n\n\n```\nvdx_helper = VDXHelper(api_url='https://vizidox-core-api.com', auth_url='https://auth.com', client_secret=secret, client_id=client_id)\n```\n\n### Mapper example\nA mapper will receive a json-formatted parameter as their input. The following example mapper will add a field\n\n```\ndef example_mapper(json_file):\n    returned_json = copy.deepcopy(json_file)\n    returned_json['additional_field'] = 'additional_value'\n    return returned_json\n```\n\n### Usage example\n\n```\nvdx_helper.upload_file(file=the_file_to_upload, mapper=example_mapper)\n```\n\n## Running the tests\n\nYou can run the tests with poetry if you like. You can also obtain the code coverage.\n\n```\npoetry run pytest --cov=vdx_helper\n```\n\n### Run the test locally with docker-compose step-by-step\n1. Spin up the docker-containers\n```\ndocker-compose up -d\n```\n\n2. Run the tests via the vdx-helper docker container\n```\ndocker-compose run vdx-helper pytest tests\n```\n\n\n## Documentation\n\nTo build the documentation locally:\n\n```shell\ncd docs\nmake html\n```\n\nThe build files can be found in docs/build. Open the generated index.html file in the html folder, and you can now \nnavigate the documentation. Repeat the above command and refresh your browser every time you update the documentation.\nAll source files are in docs/source, with vdx_helper containing the documentation generated from docstrings.\n \n## Authors\n\n* **Tiago Santos** - *Initial work* - [Vizidox](https://vizidox.com)\n* **Joana Teixeira** - *Corrections and improvements* - [Vizidox](https://vizidox.com)\n* **Rita Mariquitos** - *Corrections and improvements* - [Vizidox](https://vizidox.com)",
    'author': 'Joana Teixeira',
    'author_email': 'joana.teixeira@vizidox.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://vizidox.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime
from typing import Optional, List
from uuid import UUID


def datetime_from_string(dt: str) -> datetime:
    """
    Converts a date like string to a datetime object.

    :param dt: The string to be converted
    :type dt: str

    :return: A datetime in ISO format
    :rtype: :class:`datetime.datetime`
    """
    return datetime.fromisoformat(dt)


def optional_datetime_from_string(dt: Optional[str]) -> Optional[datetime]:
    """
    Converts a date like string to a datetime object.
    If the date is None, then returns None

    :param dt: The string to be converted
    :type dt: str, optional

    :return: A datetime in ISO format
    :rtype: :class:`datetime.datetime`, optional
    """
    return datetime_from_string(dt) if dt is not None else None


def uuid_to_string(uuid_: UUID) -> str:
    """
    Converts a UUID value to a string.

    :param uuid_: A valid UUID value
    :type uuid_: :class:`uuid.UUID`

    :return: The UUID value converted to string
    :rtype: str
    """
    return str(uuid_)


def optional_uuid_to_string(uuid_: Optional[UUID]) -> Optional[str]:
    """
    Converts a UUID value to a string, which can be None. If that is the case, then None is returned.

    :param uuid_: A valid UUID value
    :type uuid_: :class:`uuid.UUID`, optional

    :return: The UUID value converted to string
    :rtype: str, optional
    """
    return uuid_to_string(uuid_) if uuid_ is not None else None


def optional_uuids_to_string(uids: Optional[List[UUID]]) -> Optional[List[str]]:
    """
    Convert a list of valid UUIDs to string, if the list is given. If not, then return None.

    :param uids: List of UUIDs
    :type uids: List[:class:`uuid.UUID`], optional

    :return: List of the UUID in string format
    :rtype: List[str], optional
    """
    return [uuid_to_string(uid) for uid in uids] if uids is not None else None


def uuids_to_string(uids: List[UUID]) -> List[str]:
    """
    Convert a list of valid UUIDs to string.

    :param uids: List of UUIDs
    :type uids: List[:class:`uuid.UUID`]

    :return: List of the UUID in string format
    :rtype: List[str]
    """
    return [uuid_to_string(uid) for uid in uids]
